print("import: 'fsleyes'")
import fsleyes

print("import: 'fsleyes.actions'")
import fsleyes.actions

print("import: 'fsleyes.controls'")
import fsleyes.controls

print("import: 'fsleyes.displaycontext'")
import fsleyes.displaycontext

print("import: 'fsleyes.editor'")
import fsleyes.editor

print("import: 'fsleyes.gl'")
import fsleyes.gl

print("import: 'fsleyes.plotting'")
import fsleyes.plotting

print("import: 'fsleyes.plugins'")
import fsleyes.plugins

print("import: 'fsleyes.profiles'")
import fsleyes.profiles

print("import: 'fsleyes.views'")
import fsleyes.views

